/* Test Program for AUSPICE-R
 * Author: Jiaqi Tan <tanjiaqi@cmu.edu>
 * Copyright (c) 2016 Jiaqi Tan
 * 15 June 2016
 * 
 * The AUSPICE-R Test Programs are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 *
 */

#define INT_IN 3

#include <mSPI.h>
#include <rpi_defines.h>
#include <rpi.h>
#include "mFall.h"
#include <swi_functions.h>

//WIRING
//VIN 3v3 power
//3V3 <unconnected> 
//GND  GND
//CS  wiringPI PIN 10, CS
//INT1 wiringPI PIN 3, INT_IN
//INT2  <unconnected>
//SDO  wiringPI PIN 13, SPI IN (MISO)
//SDA  wiringPI PIN 12, SPI OUT (MOSI)
//SCL  wiringPI PIN 14, CLK

#ifndef MIN_SAFE_MEM
#define MIN_SAFE_MEM 0xbd00
#endif

#ifdef SAFETY
#define WORD_SIZE 4
#define ABS_MAX_SAFE_MEM 0xBF000000
#define MAX_SAFE_MEM ((unsigned int) (ABS_MAX_SAFE_MEM - WORD_SIZE))
#include <safety.h>
#endif

void updateAXL() {
    writeBatchRegister(XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT], 4);
}

void setRegisters(char THRESH_ACT,  char THRESH_INACT, char TIME_INACT, char ACT_INACT_CTL) {
#ifdef SAFETY
#define STACKSIZE 2
  register unsigned int r11_val asm ("r4");
  unsigned char *tmp_ptr;
#endif  

#ifdef SAFETY
    tmp_ptr = &(ADXL345Registers[XL345_THRESH_ACT]);
    SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = THRESH_ACT);
    tmp_ptr = &(ADXL345Registers[XL345_THRESH_INACT]);
    SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = THRESH_INACT);
    tmp_ptr = &(ADXL345Registers[XL345_TIME_INACT]);
    SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = TIME_INACT);
    tmp_ptr = &(ADXL345Registers[XL345_ACT_INACT_CTL]);
    SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = ACT_INACT_CTL);
#undef STACKSIZE
#else
    ADXL345Registers[XL345_THRESH_ACT]=THRESH_ACT;
    ADXL345Registers[XL345_THRESH_INACT]=THRESH_INACT;
    ADXL345Registers[XL345_TIME_INACT]=TIME_INACT;
    ADXL345Registers[XL345_ACT_INACT_CTL]=ACT_INACT_CTL;
#endif   
}


void action_reset() {
#ifdef SAFETY
#define STACKSIZE 4
  register unsigned int r11_val asm ("r4");
#endif

#ifdef SAFETY
  SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF0);
#undef STACKSIZE
#else
    DetectionStatus=0xF0;
#endif

    c_write(1, "F0 state\n", 9);
    setRegisters(STRIKE_THRESHOLD,  NOMOVEMENT_THRESHOLD, STABLE_TIME,XL345_INACT_Z_ENABLE | XL345_INACT_Y_ENABLE
    | XL345_INACT_X_ENABLE | XL345_INACT_AC | XL345_ACT_Z_ENABLE | XL345_ACT_Y_ENABLE | XL345_ACT_X_ENABLE |
    XL345_ACT_DC);
    //xl345Write(4, XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT]);
    updateAXL();
            
}

void action_F1_strike() {  
#ifdef SAFETY
#define STACKSIZE 3
  register unsigned int r11_val asm ("r4");
  uint32_t tmp_time;
#endif
  
    c_write(1, "F1->", 4);
#ifdef SAFETY
    SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF2);
#else
    DetectionStatus=0xF2; // Go to Status "F2"
#endif
    c_write(1, "F2 state\n", 9);
    setRegisters(STABLE_THRESHOLD,  NOMOVEMENT_THRESHOLD, STABLE_TIME,XL345_INACT_Z_ENABLE | XL345_INACT_Y_ENABLE
    | XL345_INACT_X_ENABLE | XL345_INACT_AC | XL345_ACT_Z_ENABLE | XL345_ACT_Y_ENABLE | XL345_ACT_X_ENABLE |
    XL345_ACT_AC);
    //xl345Write(4, XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT]);
    updateAXL();
    //IRQEN|=GP_TIMER_BIT; // Enable ADuC7026's Timer1 interrupt

#ifdef SAFETY
    tmp_time = getSysTicks();
    SAFETY_CHECK((&TimerWaitForStable), r11_val, STACKSIZE, TimerWaitForStable = 0);
    SAFETY_CHECK((&isTiming), r11_val, STACKSIZE, isTiming = 1);
    SAFETY_CHECK((&time), r11_val, STACKSIZE, time = tmp_time);
#undef STACKSIZE
#else
    TimerWaitForStable=0;
    isTiming = 1; //enable timer
    time = getSysTicks(); //begin timing
#endif
}

void action_F4_motion() {    
#ifdef SAFETY
#define STACKSIZE 4
  register unsigned int r11_val asm ("r4");
#endif

    c_write(1, "F4->", 4);
#ifdef SAFETY
  SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF0);
#undef STACKSIZE
#else
    DetectionStatus=0xF0; // Go to Status "F0", restart
#endif
    c_write(1, "F0 state\n", 9);
    setRegisters(STRIKE_THRESHOLD,  NOMOVEMENT_THRESHOLD, STABLE_TIME,XL345_INACT_Z_ENABLE | XL345_INACT_Y_ENABLE
    | XL345_INACT_X_ENABLE | XL345_INACT_AC | XL345_ACT_Z_ENABLE | XL345_ACT_Y_ENABLE | XL345_ACT_X_ENABLE |
    XL345_ACT_DC);
    //xl345Write(4, XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT]);
    updateAXL();
    //xl345Write(4, XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT]);
}

void action_F2() {    
#ifdef SAFETY
#define STACKSIZE 3
  register unsigned int r11_val asm ("r4");
  int tmp_val;
  unsigned char *tmp_ptr;
  unsigned short int tmp_accel1, *tmp_accel_ptr;
  unsigned long int tmp_delta_accel1, *tmp_delta_accel_ptr;
#endif
        unsigned char i;
        c_write(1, "F2->", 4);
#ifdef SAFETY
        SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF3);
#else
        DetectionStatus=0xF3; // Go to Status "F3"
#endif
        //IRQCLR = GP_TIMER_BIT;
        c_write(1, "F3 state\n", 9);
        //xl345Read(6, XL345_DATAX0, &ADXL345Registers[XL345_DATAX0]);
#ifdef SAFETY
        readBatchRegister(XL345_DATAX0,  6);

	tmp_val = regArray[0];
	tmp_ptr = &(ADXL345Registers[XL345_DATAX0]);
	SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
	tmp_val =regArray[1];
	tmp_ptr = &(ADXL345Registers[XL345_DATAX1]);
	SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
	tmp_val = regArray[2];
	tmp_ptr = &(ADXL345Registers[XL345_DATAY0]);
	SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
	tmp_val = regArray[3];
	tmp_ptr = &(ADXL345Registers[XL345_DATAY1]);
	SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
	tmp_val = regArray[4];
	tmp_ptr = &(ADXL345Registers[XL345_DATAZ0]);
	SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
	tmp_val = regArray[5];
	tmp_ptr = &(ADXL345Registers[XL345_DATAZ1]);
	SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
#else
        readBatchRegister(XL345_DATAX0,  6);
        ADXL345Registers[XL345_DATAX0] = regArray[0];
        ADXL345Registers[XL345_DATAX1] = regArray[1];
        ADXL345Registers[XL345_DATAY0] = regArray[2];
        ADXL345Registers[XL345_DATAY1] = regArray[3];
        ADXL345Registers[XL345_DATAZ0] = regArray[4];
        ADXL345Registers[XL345_DATAZ1] = regArray[5];
#endif
#ifdef SAFETY
        SAFETY_CHECK((&DeltaVectorSum), r11_val, STACKSIZE, DeltaVectorSum = 0);
#else
        DeltaVectorSum=0;
#endif        
        for(i=0;i<3; i++) 
        {
#ifdef SAFETY
	  tmp_accel1 = (ADXL345Registers[XL345_DATAX1+i*2]&0x1F);
	  tmp_accel1 = (tmp_accel1 << 8) | ADXL345Registers[XL345_DATAX0+i*2];
	  tmp_accel_ptr = &(Acceleration[i]);
	  SAFETY_CHECK(tmp_accel_ptr, r11_val, STACKSIZE, *tmp_accel_ptr = tmp_accel1);
#else
            Acceleration[i]=ADXL345Registers[XL345_DATAX1+i*2]&0x1F;
            Acceleration[i]=(Acceleration[i]<<8)|ADXL345Registers[XL345_DATAX0+i*2];
#endif
            if(Acceleration[i]<0x1000)
            {
#ifdef SAFETY
                tmp_accel_ptr = &(Acceleration[i]);
		tmp_accel1 = Acceleration[i] + 0x1000;
		SAFETY_CHECK(tmp_accel_ptr, r11_val, STACKSIZE, *tmp_accel_ptr = tmp_accel1);
#else
                Acceleration[i]=Acceleration[i]+0x1000;
#endif
            }
            else //if(Acceleration[i]>= 0x1000)
            {
#ifdef SAFETY
                tmp_accel_ptr = &(Acceleration[i]);
		tmp_accel1 = Acceleration[i] - 0x1000;
		SAFETY_CHECK(tmp_accel_ptr, r11_val, STACKSIZE, *tmp_accel_ptr = tmp_accel1);
#else
                Acceleration[i]=Acceleration[i]-0x1000;
#endif
            }
            if(Acceleration[i]>InitialStatus[i])
            {
#ifdef SAFETY
                tmp_delta_accel1 = Acceleration[i] - InitialStatus[i];
                tmp_delta_accel_ptr = &(DeltaAcceleration[i]);
                // tmp_delta_accel_ptr = (unsigned long int *)(DeltaAcceleration+i);
                SAFETY_CHECK(tmp_delta_accel_ptr, r11_val, STACKSIZE, *tmp_delta_accel_ptr = tmp_delta_accel1);
		// SAFETY_CHECK((DeltaAcceleration+i), r11_val, STACKSIZE, DeltaAcceleration[i] = tmp_delta_accel1);
#else
                DeltaAcceleration[i]=Acceleration[i]-InitialStatus[i];
#endif
            }
            else
            {
#ifdef SAFETY
                tmp_delta_accel1 = InitialStatus[i]-Acceleration[i];
                tmp_delta_accel_ptr = &(DeltaAcceleration[i]);
                // tmp_delta_accel_ptr = (unsigned long int *)(DeltaAcceleration+i);
                SAFETY_CHECK(tmp_delta_accel_ptr, r11_val, STACKSIZE, *tmp_delta_accel_ptr = tmp_delta_accel1);
		// SAFETY_CHECK((DeltaAcceleration+i), r11_val, STACKSIZE, DeltaAcceleration[i] = tmp_delta_accel1);
#else
                DeltaAcceleration[i] = InitialStatus[i]-Acceleration[i];
#endif
            }
#ifdef SAFETY
            tmp_delta_accel1 = DeltaVectorSum+DeltaAcceleration[i]*DeltaAcceleration[i];
            tmp_delta_accel_ptr = &(DeltaVectorSum);
            SAFETY_CHECK(tmp_delta_accel_ptr, r11_val, STACKSIZE, *tmp_delta_accel_ptr = tmp_delta_accel1);
#else
            DeltaVectorSum = DeltaVectorSum+DeltaAcceleration[i]*DeltaAcceleration[i];
#endif
        }
        if(DeltaVectorSum>DELTA_VECTOR_SUM_THRESHOLD) // The stable status is different from the initial status
        {
            c_write(1, "F2->", 4);
#ifdef SAFETY
            SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF4);
#else
            DetectionStatus=0xF4; // Valid fall detection
#endif
            c_write(1, "F4 state\n", 9);
            c_write(1, "ALERT FALL\n", 11);
            setRegisters(STABLE_THRESHOLD,  NOMOVEMENT_THRESHOLD, NOMOVEMENT_TIME,XL345_INACT_Z_ENABLE |
                    XL345_INACT_Y_ENABLE | XL345_INACT_X_ENABLE | XL345_INACT_AC | XL345_ACT_Z_ENABLE | XL345_ACT_Y_ENABLE |
                    XL345_ACT_X_ENABLE | XL345_ACT_AC);
            //xl345Write(4, XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT]);
            updateAXL();
        }
        else // Delta vector sum does not exceed the threshold
        {
            c_write(1, "F2->", 4);
#ifdef SAFETY
            SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF0);
#else
            DetectionStatus=0xF0; // Go to Status "F0", restart
#endif
            c_write(1, "F0 state\n", 9);
            setRegisters(STRIKE_THRESHOLD,  NOMOVEMENT_THRESHOLD,NOMOVEMENT_TIME, XL345_INACT_Z_ENABLE |
                    XL345_INACT_Y_ENABLE | XL345_INACT_X_ENABLE | XL345_INACT_AC | XL345_ACT_Z_ENABLE | XL345_ACT_Y_ENABLE |
                    XL345_ACT_X_ENABLE | XL345_ACT_DC);
            //xl345Write(4, XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT]);
            updateAXL();
        }
#ifdef SAFETY
#undef STACKSIZE
#endif
}

void action_F4_motionless() {    
#ifdef SAFETY
#define STACKSIZE 4
  register unsigned int r11_val asm ("r4");
#endif
    c_write(1, "F4->", 4);
#ifdef SAFETY
    SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF5);
#else
    DetectionStatus=0xF5; // Valid critical fall detection
#endif
    c_write(1, "F5 state\n", 9);
        //CRITICAL FALL DETECTED
    c_write(1, "ALERT CRITICAL FALL\n", 20);
    setRegisters(STRIKE_THRESHOLD,  NOMOVEMENT_THRESHOLD, STABLE_TIME,XL345_INACT_Z_ENABLE | XL345_INACT_Y_ENABLE
    | XL345_INACT_X_ENABLE | XL345_INACT_AC | XL345_ACT_Z_ENABLE | XL345_ACT_Y_ENABLE | XL345_ACT_X_ENABLE |
    XL345_ACT_DC);
    //xl345Write(4, XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT]);
    updateAXL();
    
    c_write(1, "F5->", 4);
#ifdef SAFETY
    SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF0);
#undef STACKSIZE
#else
    DetectionStatus=0xF0; // Go to Status "F0", restart
#endif
    c_write(1, "F0 state\n", 9);
}

void action_F0() {    
#ifdef SAFETY
#define STACKSIZE 3
  register unsigned int r11_val asm ("r4");
  uint32_t tmp_time;
#endif

    c_write(1, "F0->", 4);
#ifdef SAFETY
    SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF1);
#else
    DetectionStatus=0xF1; // Go to Status "F1"
#endif
    c_write(1, "F1 state\n", 9);
    setRegisters(STRIKE_THRESHOLD,  NOMOVEMENT_THRESHOLD, STABLE_TIME,XL345_INACT_Z_ENABLE | XL345_INACT_Y_ENABLE
    | XL345_INACT_X_ENABLE | XL345_INACT_AC | XL345_ACT_Z_ENABLE | XL345_ACT_Y_ENABLE | XL345_ACT_X_ENABLE |
    XL345_ACT_DC);
    //xl345Write(4, XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT]);
    updateAXL();
    
    //IRQEN|=GP_TIMER_BIT; // Enable ADuC7026's Timer1 interrupt

#ifdef SAFETY
    tmp_time = getSysTicks();
    SAFETY_CHECK((&TimerWaitForStrike), r11_val, STACKSIZE, TimerWaitForStrike = 0);
    SAFETY_CHECK((&TimerFreeFall), r11_val, STACKSIZE, TimerFreeFall = 0);
    SAFETY_CHECK((&isTiming), r11_val, STACKSIZE, isTiming = 1);
    SAFETY_CHECK((&time), r11_val, STACKSIZE, time = tmp_time);
#undef STACKSIZE
#else
    TimerWaitForStrike=0;
    TimerFreeFall=0;
    isTiming = 1; //enable timer
    time = getSysTicks(); //begin timing
#endif


}   

void action_F1_freefall() {
#ifdef SAFETY
#define STACKSIZE 3
  register unsigned int r11_val asm ("r4");
  unsigned char tmp_timer;
#endif

    if(TimerWaitForStrike<FREE_FALL_INTERVAL) // If the free fall interrupt is  continuously assert within the time of "FREE_FALL_INTERVAL",
    { // then it is considered a continuous free fall
#ifdef SAFETY
        tmp_timer = TimerFreeFall+TimerWaitForStrike;
	SAFETY_CHECK((&TimerFreeFall), r11_val, STACKSIZE, TimerFreeFall = tmp_timer);
#else
        TimerFreeFall=TimerFreeFall+TimerWaitForStrike;
#endif
    }
    else // Not a continuous free fall
    {
#ifdef SAFETY
	SAFETY_CHECK((&TimerFreeFall), r11_val, STACKSIZE, TimerFreeFall = 0);
#else
        TimerFreeFall=0;
#endif
    }
#ifdef SAFETY
    SAFETY_CHECK((&TimerWaitForStrike), r11_val, STACKSIZE, TimerWaitForStrike = 0);
#else
    TimerWaitForStrike=0;
#endif
    if(TimerFreeFall>=FREE_FALL_OVERTIME) // If the continuous time of free fall is longer than "FREE_FALL_OVERTIME"
    { // Consider that a free fall from high place is detected
        c_write(1, "F1->", 4);
#ifdef SAFETY
	SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xFF);
#else
        DetectionStatus=0xFF;
#endif
        c_write(1, "FF state\n", 9);
        c_write(1, "ALERT CONTINUOUS FALL\n", 22);
        setRegisters(STRIKE_THRESHOLD,  NOMOVEMENT_THRESHOLD, STABLE_TIME,XL345_INACT_Z_ENABLE | XL345_INACT_Y_ENABLE
    | XL345_INACT_X_ENABLE | XL345_INACT_AC | XL345_ACT_Z_ENABLE | XL345_ACT_Y_ENABLE | XL345_ACT_X_ENABLE |
    XL345_ACT_DC);
        //xl345Write(4, XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT]);
        updateAXL();
    
        c_write(1, "FF->", 4);
#ifdef SAFETY
	SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF0);
#undef STACKSIZE
#else
        DetectionStatus=0xF0;
#endif
        c_write(1, "F0 state\n", 9);
    }
    
}



void handleDetection() {    
    unsigned int duration = 0;
#ifdef SAFETY
#define STACKSIZE 3
    register unsigned int r11_val asm ("r4");
    unsigned char tmp_val;
    int tmp_int_val;
    unsigned char *tmp_ptr;
#endif
    //check for timeouts
    if(isTiming) {
        unsigned int currTime = getSysTicks();
        duration = currTime - time;
        //time = currTime;
    }
    else {
        duration = 0;
    }
    
    //if((IRQSTA & GP_TIMER_BIT)==GP_TIMER_BIT) // TIMER1 interrupt, interval 20ms 
    if(duration >= 20000) //interval 20ms
    {
        //T1CLRI = 0; // Clear TIMER1 interrupt
#ifdef SAFETY
        SAFETY_CHECK((&isTiming), r11_val, STACKSIZE, isTiming = 0);
#else
        isTiming = 0;
#endif
        duration = 0;
        if(DetectionStatus==0xF2) // Strike after weightlessness is detected, waiting for stable
        {
#ifdef SAFETY
  	    tmp_val = TimerWaitForStable + 1;
	    SAFETY_CHECK((&TimerWaitForStable), r11_val, STACKSIZE, TimerWaitForStable = tmp_val);
#else
            TimerWaitForStable++;
#endif
            if(TimerWaitForStable>=STABLE_WINDOW) // Time out, restart
            {
                //IRQCLR = GP_TIMER_BIT; // Disable ADuC7026's Timer1 interrupt
                c_write(1, "F2->", 4);
                action_reset();
            }
        }
        else if(DetectionStatus==0xF1) // Weightlessness is detected, waiting for strike
        {
#ifdef SAFETY
  	    tmp_val = TimerWaitForStrike + 1;
	    SAFETY_CHECK((&TimerWaitForStrike), r11_val, STACKSIZE, TimerWaitForStrike = tmp_val);
#else
            TimerWaitForStrike++;
#endif
	    if(TimerWaitForStrike>=STRIKE_WINDOW) // Time out, restart
            {
                //IRQCLR = GP_TIMER_BIT; // Disable ADuC7026's Timer1 interrupt
                c_write(1, "F1->", 4);
                action_reset();
            }
        }
    }   
    
    
    //state logic
    if(int0) // rising edge of int signal
    {
        //IRQCLR = SPM4_IO_BIT; // Disable ADuC7026's external interrupt
        //xl345Read(1, XL345_INT_SOURCE, &ADXL345Registers[XL345_INT_SOURCE]);
#ifdef SAFETY
      tmp_ptr = &(ADXL345Registers[XL345_INT_SOURCE]);
      tmp_int_val = readRegister(XL345_INT_SOURCE);
      SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_int_val);
#else
      ADXL345Registers[XL345_INT_SOURCE] = readRegister(XL345_INT_SOURCE);
#endif
        if((ADXL345Registers[XL345_INT_SOURCE] & XL345_ACTIVITY) == XL345_ACTIVITY) // Activity interrupt asserted
        {
            if(DetectionStatus==0xF1) // Waiting for strike, and now strike is detected
            {
                action_F1_strike();
            }
            else if(DetectionStatus==0xF4) // Waiting for long time motionless, but a movement is detected
            {
                action_F4_motion();
            }
        }
        else if((ADXL345Registers[XL345_INT_SOURCE] & XL345_INACTIVITY) == XL345_INACTIVITY) // Inactivity interrupt asserted
        {
            if(DetectionStatus==0xF2) // Waiting for stable, and now stable is detected
            {
                action_F2();
            }
            else if(DetectionStatus==0xF4) // Wait for long time motionless, and now it is detected
            {
                action_F4_motionless();
            }
        }
        else if((ADXL345Registers[XL345_INT_SOURCE] & XL345_FREEFALL) == XL345_FREEFALL) // Free fall interrupt asserted
        {
            if(DetectionStatus==0xF0) // Waiting for weightless, and now it is detected
            {
                action_F0();
            }
            else if(DetectionStatus==0xF1) // Waiting for strike after weightless, and now a new free fall is detected
            {
                action_F1_freefall();
            }
            else
            {
#ifdef SAFETY
	      SAFETY_CHECK((&TimerFreeFall), r11_val, STACKSIZE, TimerFreeFall = 0);
#undef STACKSIZE
#else
                TimerFreeFall=0;
#endif
            }
        }
        //IRQEN |=SPM4_IO_BIT; // Enable ADuC7026's external interrupt
    }
    //#ifndef SAFETY
    //    {
    //      char buf[100];
    //      atoi(DetectionStatus, buf, 99);
    //      c_write(STDOUT_FILENO, "detection status", 16);
    //      c_write(STDOUT_FILENO, buf, strlen(buf));
    //      c_write(STDOUT_FILENO, "\n", 1);
    //    }
    //#endif
}
void ADXL345_Initiate() // ADXL345 initialization, refer to ADXL345 data sheet
{
#ifdef SAFETY
#define STACKSIZE 3
    register unsigned int r11_val asm ("r4");
    unsigned char *tmp_ptr;
    int tmp_val;
#endif

#ifdef SAFETY

    tmp_val = readRegister(XL345_DEVID);
    tmp_ptr = &(ADXL345Registers[XL345_DEVID]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
    tmp_ptr = &(ADXL345Registers[XL345_OFSX]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = 0xFF);
    tmp_ptr = &(ADXL345Registers[XL345_OFSY]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = 0x05);
    tmp_ptr = &(ADXL345Registers[XL345_OFSZ]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = 0xFF);
    
    writeBatchRegister(XL345_OFSX, &ADXL345Registers[XL345_OFSX], 3);
    
    tmp_ptr = &(ADXL345Registers[XL345_THRESH_ACT]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = STRIKE_THRESHOLD);
    tmp_ptr = &(ADXL345Registers[XL345_THRESH_INACT]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = NOMOVEMENT_THRESHOLD);
    tmp_ptr = &(ADXL345Registers[XL345_TIME_INACT]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = STABLE_TIME);
    tmp_val = XL345_INACT_Z_ENABLE|XL345_INACT_Y_ENABLE | XL345_INACT_X_ENABLE
      | XL345_INACT_AC | XL345_ACT_Z_ENABLE|XL345_ACT_Y_ENABLE | XL345_ACT_X_ENABLE | XL345_ACT_DC;
    tmp_ptr = &(ADXL345Registers[XL345_ACT_INACT_CTL]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
    tmp_ptr = &(ADXL345Registers[XL345_THRESH_FF]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = FREE_FALL_THRESHOLD);
    tmp_ptr = &(ADXL345Registers[XL345_TIME_FF]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = FREE_FALL_TIME);

    writeBatchRegister(XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT], 6);
    
    tmp_ptr = &(ADXL345Registers[XL345_BW_RATE]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = XL345_RATE_100);
    tmp_ptr = &(ADXL345Registers[XL345_POWER_CTL]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = XL345_STANDBY);
    tmp_val = XL345_ACTIVITY | XL345_INACTIVITY | XL345_FREEFALL;
    tmp_ptr = &(ADXL345Registers[XL345_INT_ENABLE]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
    tmp_ptr = &(ADXL345Registers[XL345_INT_MAP]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = 0x00);
    
    writeBatchRegister(XL345_BW_RATE, &ADXL345Registers[XL345_BW_RATE],4);
    
    tmp_val = XL345_FULL_RESOLUTION | XL345_DATA_JUST_RIGHT | XL345_RANGE_16G;
    tmp_ptr = &(ADXL345Registers[XL345_DATA_FORMAT]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);

    writeRegister(XL345_DATA_FORMAT, ADXL345Registers[XL345_DATA_FORMAT]);
    
    tmp_ptr = &(ADXL345Registers[XL345_POWER_CTL]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = XL345_MEASURE);
    
    writeRegister(XL345_POWER_CTL, ADXL345Registers[XL345_POWER_CTL]);
    
    tmp_val = readRegister(XL345_INT_SOURCE);
    tmp_ptr = &(ADXL345Registers[XL345_INT_SOURCE]); SAFETY_CHECK(tmp_ptr, r11_val, STACKSIZE, *tmp_ptr = tmp_val);
#undef STACKSIZE
#else
    ADXL345Registers[XL345_DEVID] = readRegister(XL345_DEVID);
    ADXL345Registers[XL345_OFSX]=0xFF;
    ADXL345Registers[XL345_OFSY]=0x05;
    ADXL345Registers[XL345_OFSZ]=0xFF;
    
    writeBatchRegister(XL345_OFSX, &ADXL345Registers[XL345_OFSX], 3);
    
    ADXL345Registers[XL345_THRESH_ACT]=STRIKE_THRESHOLD;
    ADXL345Registers[XL345_THRESH_INACT]=NOMOVEMENT_THRESHOLD;
    ADXL345Registers[XL345_TIME_INACT]=STABLE_TIME;
    ADXL345Registers[XL345_ACT_INACT_CTL]=XL345_INACT_Z_ENABLE|XL345_INACT_Y_ENABLE | XL345_INACT_X_ENABLE
    | XL345_INACT_AC | XL345_ACT_Z_ENABLE|XL345_ACT_Y_ENABLE | XL345_ACT_X_ENABLE | XL345_ACT_DC;
    ADXL345Registers[XL345_THRESH_FF]=FREE_FALL_THRESHOLD;
    ADXL345Registers[XL345_TIME_FF]=FREE_FALL_TIME;
    
    writeBatchRegister(XL345_THRESH_ACT, &ADXL345Registers[XL345_THRESH_ACT], 6);
    
    ADXL345Registers[XL345_BW_RATE]=XL345_RATE_100;
    ADXL345Registers[XL345_POWER_CTL]=XL345_STANDBY;
    ADXL345Registers[XL345_INT_ENABLE]=XL345_ACTIVITY | XL345_INACTIVITY | XL345_FREEFALL;
    ADXL345Registers[XL345_INT_MAP]=0x00;
    
    writeBatchRegister(XL345_BW_RATE, &ADXL345Registers[XL345_BW_RATE],4);
    
    ADXL345Registers[XL345_DATA_FORMAT] = XL345_FULL_RESOLUTION | XL345_DATA_JUST_RIGHT | XL345_RANGE_16G;
    
    writeRegister(XL345_DATA_FORMAT, ADXL345Registers[XL345_DATA_FORMAT]);
    
    ADXL345Registers[XL345_POWER_CTL]=XL345_MEASURE;
    
    writeRegister(XL345_POWER_CTL, ADXL345Registers[XL345_POWER_CTL]);
    
    ADXL345Registers[XL345_INT_SOURCE] = readRegister(XL345_INT_SOURCE);
#endif

}

int main(void) {
#ifdef SAFETY
#define STACKSIZE 3
    register unsigned int r11_val asm ("r4");
    unsigned short int *tmp_status_ptr;
    int tmp_int_val;
#endif

#ifdef SAFETY
    setup_piModel1_rev2();
#else
    setup();
#endif
    initSPI(); //init SPI defaults
    ADXL345_Initiate(); // ADXL345 initialization
#ifdef SAFETY
    SAFETY_CHECK((&DetectionStatus), r11_val, STACKSIZE, DetectionStatus = 0xF0);
    tmp_status_ptr = &(InitialStatus[0]); SAFETY_CHECK(tmp_status_ptr, r11_val, STACKSIZE, *tmp_status_ptr = 0x1000);
    tmp_status_ptr = &(InitialStatus[1]); SAFETY_CHECK(tmp_status_ptr, r11_val, STACKSIZE, *tmp_status_ptr = 0x0F00);
    tmp_status_ptr = &(InitialStatus[2]); SAFETY_CHECK(tmp_status_ptr, r11_val, STACKSIZE, *tmp_status_ptr = 0x1000);
#else
    DetectionStatus=0xF0; // Clear detection status, start
    InitialStatus[0]=0x1000; // X axis=0g, unsigned short int, 13 bit resolution, 0x1000 = 4096 = 0g, +/-0xFF = +/-256 = +/-1g
    InitialStatus[1]=0x0F00; // Y axis=-1g
    InitialStatus[2]=0x1000; // Z axis=0g
#endif
    //IRQEN =SPM4_IO_BIT; // Enable ADuC7026's external interrupt, to receive the interrupt from     ADXL345 INT0
    pinMode(INT_IN, INPUT);

#ifdef SAFETY
    SAFETY_CHECK((&prev_int0), r11_val, STACKSIZE, prev_int0 = 0);
    SAFETY_CHECK((&isTiming), r11_val, STACKSIZE, isTiming = 0);
#else
    prev_int0 = 0;
    isTiming = 0;
#endif
#ifndef SAFETY
    c_write(STDOUT_FILENO, "starting detection\n", 19);
#endif
    while(1) // Endless loop, wait for interrupts
    {
        //prev_int0 = int0;
#ifdef SAFETY
        tmp_int_val = digitalRead(INT_IN); SAFETY_CHECK((&int0), r11_val, STACKSIZE, int0 = tmp_int_val);
#else
        int0 = digitalRead(INT_IN);
#endif
        handleDetection();
    }
    return 0;
}
