/* ------- Register names ------- */
#define XL345_DEVID 0x00
#define XL345_RESERVED1 0x01
#define XL345_THRESH_TAP 0x1d
#define XL345_OFSX 0x1e
#define XL345_OFSY 0x1f
#define XL345_OFSZ 0x20
#define XL345_DUR 0x21
#define XL345_LATENT 0x22
#define XL345_WINDOW 0x23
#define XL345_THRESH_ACT 0x24
#define XL345_THRESH_INACT 0x25
#define XL345_TIME_INACT 0x26
#define XL345_ACT_INACT_CTL 0x27
#define XL345_THRESH_FF 0x28
#define XL345_TIME_FF 0x29
#define XL345_TAP_AXES 0x2a
#define XL345_ACT_TAP_STATUS 0x2b
#define XL345_BW_RATE 0x2c
#define XL345_POWER_CTL 0x2d
#define XL345_INT_ENABLE 0x2e
#define XL345_INT_MAP 0x2f
#define XL345_INT_SOURCE 0x30
#define XL345_DATA_FORMAT 0x31
#define XL345_DATAX0 0x32
#define XL345_DATAX1 0x33
#define XL345_DATAY0 0x34
#define XL345_DATAY1 0x35
#define XL345_DATAZ0 0x36
#define XL345_DATAZ1 0x37
#define XL345_FIFO_CTL 0x38
#define XL345_FIFO_STATUS 0x39

/* Bit values in ACT_INACT_CTL */
#define XL345_INACT_Z_ENABLE 0x01
#define XL345_INACT_Z_DISABLE 0x00
#define XL345_INACT_Y_ENABLE 0x02
#define XL345_INACT_Y_DISABLE 0x00
#define XL345_INACT_X_ENABLE 0x04
#define XL345_INACT_X_DISABLE 0x00
#define XL345_INACT_AC 0x08
#define XL345_INACT_DC 0x00
#define XL345_ACT_Z_ENABLE 0x10
#define XL345_ACT_Z_DISABLE 0x00
#define XL345_ACT_Y_ENABLE 0x20
#define XL345_ACT_Y_DISABLE 0x00
#define XL345_ACT_X_ENABLE 0x40
#define XL345_ACT_X_DISABLE 0x00
#define XL345_ACT_AC 0x80
#define XL345_ACT_DC 0x00

#define XL345_THRESH_ACT 0x24
#define XL345_THRESH_INACT 0x25
#define XL345_TIME_INACT 0x26
#define XL345_ACT_INACT_CTL 0x27
#define XL345_THRESH_FF 0x28
#define XL345_TIME_FF 0x29

#define XL345_FREEFALL 0x04
#define XL345_INACTIVITY 0x08
#define XL345_ACTIVITY 0x10

#define XL345_RATE_100 0x0a

/* Bit values in POWER_CTL */
#define XL345_WAKEUP_8HZ 0x00
#define XL345_WAKEUP_4HZ 0x01
#define XL345_WAKEUP_2HZ 0x02
#define XL345_WAKEUP_1HZ 0x03
#define XL345_SLEEP 0x04
#define XL345_MEASURE 0x08
#define XL345_STANDBY 0x00
#define XL345_AUTO_SLEEP 0x10
#define XL345_ACT_INACT_SERIAL 0x20
#define XL345_ACT_INACT_CONCURRENT 0x00

/* Bit values in INT_ENABLE, INT_MAP, and INT_SOURCE are identical.
Use these bit values to read or write any of these registers. */
#define XL345_OVERRUN 0x01
#define XL345_WATERMARK 0x02
#define XL345_FREEFALL 0x04
#define XL345_INACTIVITY 0x08
#define XL345_ACTIVITY 0x10
#define XL345_DOUBLETAP 0x20
#define XL345_SINGLETAP 0x40
#define XL345_DATAREADY 0x80

/* Bit values in DATA_FORMAT */
/* Register values read in DATAX0 through DATAZ1 are dependent on the
value specified in data format. Customer code will need to interpret
the data as desired. */
#define XL345_RANGE_2G 0x00
#define XL345_RANGE_4G 0x01
#define XL345_RANGE_8G 0x02
#define XL345_RANGE_16G 0x03
#define XL345_DATA_JUST_RIGHT 0x00
#define XL345_DATA_JUST_LEFT 0x04
#define XL345_10BIT 0x00
#define XL345_FULL_RESOLUTION 0x08
#define XL345_INT_LOW 0x20
#define XL345_INT_HIGH 0x00
#define XL345_SPI3WIRE 0x40
#define XL345_SPI4WIRE 0x00
#define XL345_SELFTEST 0x80

// Definitions used for Fall Detection Algorithm
#define STRIKE_THRESHOLD 0x20 //62.5mg/LSB, 0x20=2g
#define STRIKE_WINDOW 0x0A //20ms/LSB, 0x0A=10=200ms
#define STABLE_THRESHOLD 0x08 //62.5mg/LSB, 0x10=0.5g
#define STABLE_TIME 0x02 //1s/LSB, 0x02=2s
#define STABLE_WINDOW 0xAF //20ms/LSB, 0xAF=175=3.5s
#define NOMOVEMENT_THRESHOLD 0x03 //62.5mg/LSB, 0x03=0.1875g
#define NOMOVEMENT_TIME 0x0A //1s/LSB, 0x0A=10s
#define FREE_FALL_THRESHOLD 0x0C //62.5mg/LSB, 0x0C=0.75g
#define FREE_FALL_TIME 0x06 //5ms/LSB, 0x06=30ms
#define FREE_FALL_OVERTIME 0x0F //20ms/LSB, 0x0F=15=300ms
#define FREE_FALL_INTERVAL 0x05 //20ms/LSB, 0x05=100ms
#define DELTA_VECTOR_SUM_THRESHOLD 0x7D70 //1g=0xFF, 0x7D70=0.7g^2


// Variables used for Fall Detection Algorithm
unsigned int time;
int isTiming;

 //variables for simulating interrupt
int int0;
int prev_int0;

unsigned char ADXL345Registers[57];
unsigned char DetectionStatus; // Detection status:
// 0xF0: Start
// 0xF1: Weightlessness
// 0xF2: Strike after weightlessness
// 0xF3: Stable after strike, valid fall detection
// 0xF4: Long time motionless, valid critical fall detection
// 0xFF: Continuous free fall, free fall from a high place
unsigned char TimerWaitForStable; // Counter of time that wait for stable after strike
unsigned char TimerWaitForStrike; // Counter of time that wait for strike after weightless
unsigned char TimerFreeFall; // Counter of continuous time for free fall
unsigned short int InitialStatus[3]; // Initial status for X-, Y-, Z- axis
unsigned short int Acceleration[3]; // Acceleration for X-, Y-, Z- axis
unsigned long int DeltaAcceleration[3]; // Acceleration[] - Initial_Status[]
unsigned long int DeltaVectorSum; // Vector sum of the DeltaAcceleration[]

void handleDetection();
void ADXL345_Initiate();
void action_F4_motion();
void action_F4_motionless();
void action_reset();
void action_F1_strike();
void action_F2();
void action_F1_freefall();
void updateAXL();
void setRegisters(char THESH_ACT,  char THRESH_INACT, char TIME_INACT, char ACT_INACT_CTL);