/* Test Program for AUSPICE-R
 * Authors: Jiaqi Tan <tanjiaqi@cmu.edu>, Hui Jun Tay <htay@andrew.cmu.edu>
 * Modifications Copyright (c) 2016 Jiaqi Tan, Hui Jun Tay
 * 15 June 2016
 * 
 * The AUSPICE-R Test Programs are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * Originally part of the MiBench benchmark suite (http://vhosts.eecs.umich.edu/mibench/)
 */

/* +++Date last modified: 05-Jul-1997 */

/*
**        A Pratt-Boyer-Moore string search, written by Jerry Coffin
**  sometime or other in 1991.  Removed from original program, and
**  (incorrectly) rewritten for separate, generic use in early 1992.
**  Corrected with help from Thad Smith, late March and early
**  April 1992...hopefully it's correct this time. Revised by Bob Stout.
**
**  This is hereby placed in the Public Domain by its author.
**
**  10/21/93 rdg  Fixed bug found by Jeff Dunlop
*/

#define _PRINT
#define _SAFETYCHECK

#ifdef PRINT
#include <stdio.h>
#endif // PRINT

#include <stddef.h>
#include <string.h>
#include <limits.h>

#ifdef SAFETY
#define MIN_SAFE_MEM 0x8838
#define WORD_SIZE 4
#define DOUBLE_WORD_SIZE 8
#define ABS_MAX_SAFE_MEM 0xBF000000
#define MAX_SAFE_MEM ((unsigned int) (ABS_MAX_SAFE_MEM - WORD_SIZE))
#include <safety.h>
#endif // SAFETYCHECK

static size_t table[UCHAR_MAX + 1];
static size_t len;
static char *findme;

/*
**  Call this with the string to locate to initialize the table
*/

#include "input.h"

void init_search(const char *string)
{
  size_t i, j;
  char *c;
#ifdef SAFETY
  register unsigned int r11_val asm ("r4");
#endif  

  j = strlen(string);
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if ((unsigned int) (&len) < (r11_val - (4 * WORD_SIZE)) &&
      (r11_val >= (4 * WORD_SIZE))) {
#endif
    len = j;
#ifdef SAFETY
  }
#endif
	  
  for (i = 0; i <= UCHAR_MAX; i++) {                      /* rdg 10/93 */
#ifdef SAFETY
	  SAFETY_CHECK((i+table), r11_val, 4, table[i] = len;);
#else
    table[i] = len;
#endif
  }
  
  for (i = 0; i < len; i++) {
    c = &(table[(unsigned char)string[i]]);
#ifdef SAFETY
	SAFETY_CHECK(c, r11_val, 4, *c = len - i - 1);
#else      
	*c = len - i - 1;
#endif
  }

#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int) &findme < (r11_val - (4 * WORD_SIZE)))&&
      ((r11_val >= (4*WORD_SIZE)))) {
#endif
	findme = (char *)string;
#ifdef SAFETY
  }
#endif
}

/*
**  Call this with a buffer to search
*/

char *strsearch(const char *string)
{
      register size_t shift;
      register size_t pos = len - 1;
      char *here;
      size_t limit=strlen(string);

      while (pos < limit)
      {
            while( pos < limit &&
                  (shift = table[(unsigned char)string[pos]]) > 0)
            {
                  pos += shift;
            }
            if (0 == shift)
            {
                  if (0 == strncmp(findme,
                        here = (char *)&string[pos-len+1], len))
                  {
                        return(here);
                  }
                  else  pos++;
            }
      }
      return NULL;
}

int main()
{
      char *here;
      int i;
      int counter = 0;
      for (i = 0; find_strings[i]; i++)
      {
            init_search(find_strings[i]);
            here = strsearch(search_strings[i]);
#ifdef PRINT
            printf("\"%s\" is%s in \"%s\"", find_strings[i],
                  here ? "" : " not", search_strings[i]);
#endif //print
            if (here) {
                counter = counter + 1;
#ifdef PRINT
		printf(" [\"%s\"]", here);
		putchar('\n');
#endif // PRINT
	    }
      }

      return counter;
}
