/* Test Program for AUSPICE-R
 * Authors: Jiaqi Tan <tanjiaqi@cmu.edu>, Hui Jun Tay <htay@andrew.cmu.edu>
 * Modifications Copyright (c) 2016 Jiaqi Tan, Hui Jun Tay
 * 15 June 2016
 * 
 * The AUSPICE-R Test Programs are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * Originally part of the WiringPi library (http://wiringpi.com)
 * Modified to add safety-checks for AUSPICE and AUSPICE-R
 */

/*
 * wiringPi:
 *	Arduino compatable (ish) Wiring library for the Raspberry Pi
 *	Copyright (c) 2012 Gordon Henderson
 ***********************************************************************
 * This file is part of wiringPi:
 *	https://projects.drogon.net/raspberry-pi/wiringpi/
 *
 *    wiringPi is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    wiringPi is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with wiringPi.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************
 */

#include <swi_functions.h>
#include <rpi.h>
#include <rpi_defines.h>

#ifndef MIN_SAFE_MEM
#define MIN_SAFE_MEM 0xaa00
#endif

#define WORD_SIZE 4
#define ABS_MAX_SAFE_MEM 0xBF000000
#define MAX_SAFE_MEM ((unsigned int) (ABS_MAX_SAFE_MEM - WORD_SIZE))

#define GET_FRAME_POINTER(dest_var)		\
  asm ( "mov r4,r11"				\
	: "=r" (dest_var)			\
	: /* no inputs */			\
	: /* no clobber */)			

#define SAFETY_CHECK(addr, fp_var, stack_size, op)  GET_FRAME_POINTER(r11_val);	\
  if (((unsigned int)(addr) >= MIN_SAFE_MEM) &&					\
      ((unsigned int)(addr) <= MAX_SAFE_MEM) &&					\
      ((unsigned int)(addr) < ((fp_var)-((stack_size)*WORD_SIZE))) &&			\
      ((fp_var) >= ((stack_size)*WORD_SIZE))) { op; } 


#ifndef	TRUE
#  define	TRUE	(1==1)
#  define	FALSE	(1==2)
#endif

#define MAX_LCDS 8

// HD44780U Commands

#define	LCD_CLEAR	0x01
#define	LCD_HOME	0x02
#define	LCD_ENTRY	0x04
#define	LCD_CTRL	0x08
#define	LCD_CDSHIFT	0x10
#define	LCD_FUNC	0x20
#define	LCD_CGRAM	0x40
#define	LCD_DGRAM	0x80

// Bits in the entry register

#define	LCD_ENTRY_SH		0x01
#define	LCD_ENTRY_ID		0x02

// Bits in the control register

#define	LCD_BLINK_CTRL		0x01
#define	LCD_CURSOR_CTRL		0x02
#define	LCD_DISPLAY_CTRL	0x04

// Bits in the function register

#define	LCD_FUNC_F	0x04
#define	LCD_FUNC_N	0x08
#define	LCD_FUNC_DL	0x10

#define	LCD_CDSHIFT_RL	0x04

struct lcdDataStruct
{
  int bits, rows, cols ;
  int rsPin, strbPin ;
  int dataPins [8] ;
  int cx, cy ;
} ;

struct lcdDataStruct *lcds [MAX_LCDS] ;

static int lcdControl ;

// Row offsets

static const int rowOff [4] = { 0x00, 0x40, 0x14, 0x54 } ;


/*
 * strobe:
 *	Toggle the strobe (Really the "E") pin to the device.
 *	According to the docs, data is latched on the falling edge.
 *********************************************************************************
 */


static void strobe (const struct lcdDataStruct *lcd)
{
  // brute-force calling of c_nanosleep to reduce calls to delayMicroseconds
  struct timespec sleeper50ms ;
  unsigned int howLong = 50;
  unsigned int uSecs = howLong % 1000000 ;
  unsigned int wSecs = howLong / 1000000 ;
  sleeper50ms.tv_sec  = wSecs ;
  sleeper50ms.tv_nsec = (long)(uSecs * 1000L) ;

// Note timing changes for new version of delayMicroseconds ()
  digitalWrite (lcd->strbPin, 1) ;   c_nanosleep (&sleeper50ms, 0) ;
  digitalWrite (lcd->strbPin, 0) ;   c_nanosleep (&sleeper50ms, 0) ;
}

/*
 * sentDataCmd:
 *	Send an data or command byte to the display.
 *********************************************************************************
 */
static void sendDataCmd (const struct lcdDataStruct *lcd, unsigned char data)
{
  register unsigned char myData = data ;
  unsigned char          i, d4, d5_8, j ;
  int tmp_write_val;

  //  if (lcd->bits == 4)
  //  {
    d4 = (myData >> 4) & 0x0F;
    d5_8 = myData & 0x0F ;
    for (i = 0 ; i < 8 ; ++i)
    {
      if (i < 4) tmp_write_val = d4 & 1;
      else tmp_write_val = d5_8 & 1;

      if (i < 4) j = i;
      else j = (i % 4);
      digitalWrite (lcd->dataPins [j], tmp_write_val) ;

      if (i < 4) d4 >>= 1 ;
      else d5_8 >>= 1;

      if (i == 3) strobe (lcd) ;
    }
    // strobe (lcd) ;

    // d4 = myData & 0x0F ;
    // for (i = 0 ; i < 4 ; ++i)
    // {
    //  digitalWrite (lcd->dataPins [i], (d4 & 1)) ;
    //  d4 >>= 1 ;
    //}
    // we are not using 8-bit mode for now
    /*  }
  else
  {
    for (i = 0 ; i < 8 ; ++i)
    {
      digitalWrite (lcd->dataPins [i], (myData & 1)) ;
      myData >>= 1 ;
    }
    }*/
  strobe (lcd) ;
}

/*
 * putCommand:
 *	Send a command byte to the display
 *********************************************************************************
 */

#define putCommand(lcd, command) \
digitalWrite(lcd->rsPin, 0) ; \
sendDataCmd(lcd, command) ; \
delay (2) ;

/*
static void put4Command (const struct lcdDataStruct *lcd, unsigned char command)
{
  register unsigned char myCommand = command ;
  register unsigned char i ;

  digitalWrite (lcd->rsPin,   0) ;

  for (i = 0 ; i < 4 ; ++i)
  {
    digitalWrite (lcd->dataPins [i], (myCommand & 1)) ;
    myCommand >>= 1 ;
  }
  strobe (lcd) ;
}
*/

#define put4Command(lcd, command) \
{ \
  register unsigned char myCommand = command ; \
  register unsigned char i ; \
  digitalWrite(lcd->rsPin,   0) ; \
  for (i = 0 ; i < 4 ; ++i) \
  { \
    digitalWrite(lcd->dataPins [i], (myCommand & 1)) ; \
    myCommand >>= 1 ; \
  } \
  strobe(lcd); \
}

/*
 *********************************************************************************
 * User Callable code below here
 *********************************************************************************
 */

/*
 * lcdHome: lcdClear:
 *	Home the cursor or clear the screen.
 *********************************************************************************
 */
/*
void lcdHome (const int fd)
{
  register unsigned int r11_val asm ("r4");
  struct lcdDataStruct *lcd = lcds [fd] ;

  putCommand (lcd, LCD_HOME);

  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&(lcd->cx)) >= MIN_SAFE_MEM) && 
      ((unsigned int)(&(lcd->cx)) <= MAX_SAFE_MEM) && 
      ((unsigned int)(&(lcd->cx)) < (r11_val-(3*WORD_SIZE))) && 
      (r11_val >= (3*WORD_SIZE))) {
    lcd->cx = 0;
  }

  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&(lcd->cy)) >= MIN_SAFE_MEM) && 
      ((unsigned int)(&(lcd->cy)) <= MAX_SAFE_MEM) && 
      ((unsigned int)(&(lcd->cy)) < (r11_val-(3*WORD_SIZE))) && 
      (r11_val >= (3*WORD_SIZE))) {
    lcd->cy = 0;
  }

  delay (5) ;
}
*/

void lcdClear (const int fd)
{
  register unsigned int r11_val asm ("r4");
  struct lcdDataStruct *lcd = lcds [fd] ;
  int *tmp_ptr;

  putCommand (lcd, LCD_CLEAR) ;
  putCommand (lcd, LCD_HOME) ;

  tmp_ptr = &(lcd->cx);
  SAFETY_CHECK(tmp_ptr, r11_val, 3, (*tmp_ptr) = 0);

  tmp_ptr = &(lcd->cy);
  SAFETY_CHECK(tmp_ptr, r11_val, 3, (*tmp_ptr) = 0);

  delay (5) ;
}


/*
 * lcdDisplay: lcdCursor: lcdCursorBlink:
 *	Turn the display, cursor, cursor blinking on/off
 *********************************************************************************
 */


void lcdDisplayCursorBlink (const int fd, int state, int cmd)
{
  register unsigned int r11_val asm ("r4");
  struct lcdDataStruct *lcd = lcds [fd] ;

  if (state) {
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int)(&(lcdControl)) >= MIN_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) <= MAX_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) < (r11_val-(3*WORD_SIZE))) && 
	(r11_val >= (3*WORD_SIZE))) {
      lcdControl = lcdControl | /*LCD_DISPLAY_CTRL*/ cmd;
    }
  } else {
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int)(&(lcdControl)) >= MIN_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) <= MAX_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) < (r11_val-(3*WORD_SIZE))) && 
	(r11_val >= (3*WORD_SIZE))) {
      lcdControl = lcdControl & /*(~LCD_DISPLAY_CTRL)*/ (~cmd);
    }
  }

  putCommand (lcd, LCD_CTRL | lcdControl) ; 
}
/*
void lcdDisplay (const int fd, int state)
{
  register unsigned int r11_val asm ("r4");
  struct lcdDataStruct *lcd = lcds [fd] ;

  if (state) {
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int)(&(lcdControl)) >= MIN_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) <= MAX_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) < (r11_val-(3*WORD_SIZE))) && 
	(r11_val >= (3*WORD_SIZE))) {
      lcdControl = lcdControl | LCD_DISPLAY_CTRL;
    }
  } else {
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int)(&(lcdControl)) >= MIN_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) <= MAX_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) < (r11_val-(3*WORD_SIZE))) && 
	(r11_val >= (3*WORD_SIZE))) {
      lcdControl = lcdControl & (~LCD_DISPLAY_CTRL);
    }
  }

  putCommand (lcd, LCD_CTRL | lcdControl) ; 
}

void lcdCursor (const int fd, int state)
{
  register unsigned int r11_val asm ("r4");
  struct lcdDataStruct *lcd = lcds [fd] ;

  if (state) {
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int)(&(lcdControl)) >= MIN_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) <= MAX_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) < (r11_val-(3*WORD_SIZE))) && 
	(r11_val >= (3*WORD_SIZE))) {
      lcdControl = lcdControl | LCD_CURSOR_CTRL;
    }
  } else {
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int)(&(lcdControl)) >= MIN_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) <= MAX_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) < (r11_val-(3*WORD_SIZE))) && 
	(r11_val >= (3*WORD_SIZE))) {
      lcdControl = lcdControl & (~LCD_CURSOR_CTRL);
    }
  }

  putCommand (lcd, LCD_CTRL | lcdControl) ; 
}

void lcdCursorBlink (const int fd, int state)
{
  register unsigned int r11_val asm ("r4");
  struct lcdDataStruct *lcd = lcds [fd] ;

  if (state) {
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int)(&(lcdControl)) >= MIN_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) <= MAX_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) < (r11_val-(3*WORD_SIZE))) && 
	(r11_val >= (3*WORD_SIZE))) {
      lcdControl = lcdControl | LCD_BLINK_CTRL ;
    }
  } else {
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int)(&(lcdControl)) >= MIN_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) <= MAX_SAFE_MEM) && 
	((unsigned int)(&(lcdControl)) < (r11_val-(3*WORD_SIZE))) && 
	(r11_val >= (3*WORD_SIZE))) {
      lcdControl = lcdControl & (~LCD_BLINK_CTRL) ;
    }
  }

  putCommand (lcd, LCD_CTRL | lcdControl) ; 
}
*/

/*
 * lcdSendCommand:
 *	Send any arbitary command to the display
 *********************************************************************************
 */

void lcdSendCommand (const int fd, unsigned char command)
{
  struct lcdDataStruct *lcd = lcds [fd] ;
  putCommand (lcd, command) ;
}


/*
 * lcdPosition:
 *	Update the position of the cursor on the display.
 *	Ignore invalid locations.
 *********************************************************************************
 */

void lcdPosition (const int fd, int x, int y)
{
  register unsigned int r11_val asm ("r4");
  struct lcdDataStruct *lcd = lcds [fd] ;

  if ((x > lcd->cols) || (x < 0))
    return ;
  if ((y > lcd->rows) || (y < 0))
    return ;

  putCommand (lcd, x + (LCD_DGRAM | rowOff [y])) ;

  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&(lcd->cx)) >= MIN_SAFE_MEM) && 
      ((unsigned int)(&(lcd->cx)) <= MAX_SAFE_MEM) && 
      ((unsigned int)(&(lcd->cx)) < (r11_val-(3*WORD_SIZE))) && 
      (r11_val >= (3*WORD_SIZE))) {
    lcd->cx = x ;
  }
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&(lcd->cy)) >= MIN_SAFE_MEM) && 
      ((unsigned int)(&(lcd->cy)) <= MAX_SAFE_MEM) && 
      ((unsigned int)(&(lcd->cy)) < (r11_val-(3*WORD_SIZE))) && 
      (r11_val >= (3*WORD_SIZE))) {
    lcd->cy = y ;
  }
}


/*
 * lcdCharDef:
 *	Defines a new character in the CGRAM
 *********************************************************************************
 */

void lcdCharDef (const int fd, int index, unsigned char data [8])
{
  struct lcdDataStruct *lcd = lcds [fd] ;
  int i ;

  putCommand (lcd, LCD_CGRAM | ((index & 7) << 3)) ;

  digitalWrite (lcd->rsPin, 1) ;
  for (i = 0 ; i < 8 ; ++i)
    sendDataCmd (lcd, data [i]) ;
}


/*
 * lcdPutchar:
 *	Send a data byte to be displayed on the display. We implement a very
 *	simple terminal here - with line wrapping, but no scrolling. Yet.
 *********************************************************************************
 */

void lcdPutchar (const int fd, unsigned char data)
{
  register unsigned int r11_val asm ("r4");
  struct lcdDataStruct *lcd = lcds [fd] ;
  int *tmp_ptr;

  digitalWrite (lcd->rsPin, 1) ;
  sendDataCmd  (lcd, data) ;

  tmp_ptr = &(lcd->cx);
  SAFETY_CHECK(tmp_ptr, r11_val, 3, (*tmp_ptr)++);

  if (lcd->cx == lcd->cols)
  {
    
    tmp_ptr = &(lcd->cx);
    SAFETY_CHECK(tmp_ptr, r11_val, 3, (*tmp_ptr) = 0);

    tmp_ptr = &(lcd->cy);
    SAFETY_CHECK(tmp_ptr, r11_val, 3, (*tmp_ptr)++);

    if (lcd->cy == lcd->rows) {
      tmp_ptr = &(lcd->cy);
      SAFETY_CHECK(tmp_ptr, r11_val, 3, (*tmp_ptr) = 0);
    }
    
    putCommand (lcd, lcd->cx + (LCD_DGRAM | rowOff [lcd->cy])) ;
  }
}


/*
 * lcdPuts:
 *	Send a string to be displayed on the display
 *********************************************************************************
 */

void lcdPuts (const int fd, const char *string)
{
  while (*string)
    lcdPutchar (fd, *string++) ;
}

struct lcdDataStruct LCD;

/*
 * lcdPrintf:
 *	Printf to an LCD display
 *********************************************************************************
 */

/*
 * lcdInit:
 *	Take a lot of parameters and initialise the LCD, and return a handle to
 *	that LCD, or -1 if any error.
 *********************************************************************************
 */
static int initialised = 0;

int lcdInit (const int rows, const int cols, const int bits,
	const int rs, const int strb,
	const int d0, const int d1, const int d2, const int d3, const int d4,
	const int d5, const int d6, const int d7)
{
  register unsigned int r11_val asm ("r4");
  unsigned char func ;
  int i ;
  int lcdFd = -1 ;
  struct lcdDataStruct *lcd = & LCD, **lcd_ptr;
  int *tmp_ptr;

  if (initialised == 0)
  {
    SAFETY_CHECK(&initialised, r11_val, 4, initialised = 1);

    for (i = 0 ; i < MAX_LCDS ; ++i) {
      SAFETY_CHECK((lcds+i), r11_val, 4, lcds[i] = 0);
    }
  }

// Simple sanity checks

  if (! ((bits == 4) || (bits == 8)))
    return -1 ;

  if ((rows < 0) || (rows > 20))
    return -1 ;

  if ((cols < 0) || (cols > 20))
    return -1 ;

// Create a new LCD:

  for (i = 0 ; i < MAX_LCDS ; ++i)
  {
    if (lcds [i] == 0)
    {
      lcdFd = i ;
      break ;
    }
  }

  if (lcdFd == -1)
    return -1 ;

  // lcd = (struct lcdDataStruct *)malloc (sizeof (struct lcdDataStruct)) ;
  if (lcd == 0)
    return -1 ;

  tmp_ptr = &(lcd->rsPin);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = rs);
  tmp_ptr = &(lcd->strbPin);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = strb);
  tmp_ptr = &(lcd->bits);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = 4);
  tmp_ptr = &(lcd->rows);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = rows);
  tmp_ptr = &(lcd->cols);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = cols);
  tmp_ptr = &(lcd->cx);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = 0);
  tmp_ptr = &(lcd->cy);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = 0);

  tmp_ptr = &(lcd->dataPins[0]);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = d0);
  tmp_ptr = &(lcd->dataPins[1]);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = d1);
  tmp_ptr = &(lcd->dataPins[2]);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = d2);
  tmp_ptr = &(lcd->dataPins[3]);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = d3);
  tmp_ptr = &(lcd->dataPins[4]);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = d4);
  tmp_ptr = &(lcd->dataPins[5]);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = d5);
  tmp_ptr = &(lcd->dataPins[6]);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = d6);
  tmp_ptr = &(lcd->dataPins[7]);
  SAFETY_CHECK(tmp_ptr, r11_val, 4, *(tmp_ptr) = d7);

  // lcds [lcdFd] = lcd ;
  lcd_ptr = &(lcds[lcdFd]);
  SAFETY_CHECK(lcd_ptr, r11_val, 4, *(lcd_ptr) = lcd );

  digitalWrite (lcd->rsPin,   0) ; pinMode (lcd->rsPin,   OUTPUT) ;
  digitalWrite (lcd->strbPin, 0) ; pinMode (lcd->strbPin, OUTPUT) ;

  for (i = 0 ; i < bits ; ++i)
  {
    digitalWrite (lcd->dataPins [i], 0) ;
    pinMode      (lcd->dataPins [i], OUTPUT) ;
  }
  delay (35) ; // mS


// 4-bit mode?
//	OK. This is a PIG and it's not at all obvious from the documentation I had,
//	so I guess some others have worked through either with better documentation
//	or more trial and error... Anyway here goes:
//
//	It seems that the controller needs to see the FUNC command at least 3 times
//	consecutively - in 8-bit mode. If you're only using 8-bit mode, then it appears
//	that you can get away with one func-set, however I'd not rely on it...
//
//	So to set 4-bit mode, you need to send the commands one nibble at a time,
//	the same three times, but send the command to set it into 8-bit mode those
//	three times, then send a final 4th command to set it into 4-bit mode, and only
//	then can you flip the switch for the rest of the library to work in 4-bit
//	mode which sends the commands as 2 x 4-bit values.

//  if (bits == 4)
//  {
    func = LCD_FUNC | LCD_FUNC_DL ;			// Set 8-bit mode 3 times
    for (i = 0; i < 3; i++) {
      put4Command (lcd, func >> 4) ; delay (35) ;
    }
    //put4Command (lcd, func >> 4) ; delay (35) ;
    //put4Command (lcd, func >> 4) ; delay (35) ;
    func = LCD_FUNC ;					// 4th set: 4-bit mode
    put4Command (lcd, func >> 4) ; delay (35) ;
    SAFETY_CHECK(&(lcd->bits), r11_val, 4, lcd->bits = 4 );
    //  }
    /*  else
  {
    func = LCD_FUNC | LCD_FUNC_DL ;
    putCommand  (lcd, func     ) ; delay (35) ;
    putCommand  (lcd, func     ) ; delay (35) ;
    putCommand  (lcd, func     ) ; delay (35) ;
  }
    */

  if (lcd->rows > 1)
  {
    func |= LCD_FUNC_N ;
    putCommand (lcd, func) ; delay (35) ;
  }

// Rest of the initialisation sequence

  lcdDisplayCursorBlink (lcdFd, TRUE, LCD_DISPLAY_CTRL) ;
  //lcdDisplayCursorBlink (lcdFd, FALSE, LCD_CURSOR_CTRL) ;
  //lcdDisplayCursorBlink (lcdFd, FALSE, LCD_BLINK_CTRL) ;
  lcdClear       (lcdFd) ;

  putCommand (lcd, LCD_ENTRY   | LCD_ENTRY_ID) ;
  putCommand (lcd, LCD_CDSHIFT | LCD_CDSHIFT_RL) ;

  return lcdFd ;
}
