/* Test Program for AUSPICE-R
 * Authors: Jiaqi Tan <tanjiaqi@cmu.edu>, Hui Jun Tay <htay@andrew.cmu.edu>
 * Copyright (c) 2016 Jiaqi Tan, Hui Jun Tay
 * 15 June 2016
 * 
 * The AUSPICE-R Test Programs are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 *
 */

//references
//http://raspberrypi-aa.github.io/session3/spi.html
//http://www.analog.com/media/en/technical-documentation/data-sheets/ADXL345.PDF 
//https://www.sparkfun.com/tutorials/240

#include <rpi.h>
#include <rpi_defines.h>
#include <mSPI.h>
//#include "rpi.h"
//#include "rpi_defines.h"
//#include "mSPI.h"

#ifndef MIN_SAFE_MEM
#define MIN_SAFE_MEM 0xbd00
#endif

#ifdef SAFETY
#define WORD_SIZE 4
#define ABS_MAX_SAFE_MEM 0xBF000000
#define MAX_SAFE_MEM ((unsigned int) (ABS_MAX_SAFE_MEM - WORD_SIZE))
#define GET_FRAME_POINTER(dest_var)		\
  asm ( "mov r4,r11"				\
	: "=r" (dest_var)			\
	: /* no inputs */			\
	: /* no clobber */)			
#define SAFETY_CHECK(addr, fp_var, stack_size, op)  GET_FRAME_POINTER(r11_val);	\
  if (((unsigned int)(addr) >= MIN_SAFE_MEM) &&					\
      ((unsigned int)(addr) <= MAX_SAFE_MEM) &&					\
      ((unsigned int)(addr) < ((fp_var)-((stack_size)*WORD_SIZE))) &&			\
      ((fp_var) >= ((stack_size)*WORD_SIZE))) { op; } 
#endif

void writeBatchRegister(char registerAddress, char* valuePtr, int length){  
  //Set the Chip select pin low to start an SPI packet.
  //printf("set CS LOW\n");		
  digitalWrite(CS, LOW);
  digitalWrite(SPI_CLK, LOW);
  
  //write WRITE - high
  char packet = 0x7f & registerAddress;
  
  //set MBL - high
  packet = 0x40 | registerAddress;
  
  //transfer WRITE + MB + address
  int i = 0;
  for (i = 0; i < 8; i++) {
    if (packet & 0x80) {

      digitalWrite(SPI_OUT, HIGH);
    }
    else {

      digitalWrite(SPI_OUT, LOW);
    }
    
    packet <<= 1;
    digitalWrite(SPI_CLK, HIGH);
    digitalWrite(SPI_CLK, LOW);
  }
    

  //transfer data

  int j = 0;
  i = 0;
  for(j = 0; j < length; j++) { 
      char value = valuePtr[j];
      for (i = 0; i < 8; i++) {
            
        if (value & 0x80) {

          digitalWrite(SPI_OUT, HIGH);
        }
        else {

          digitalWrite(SPI_OUT, LOW);
        }
        
        value <<= 1;
        digitalWrite(SPI_CLK, HIGH);
        digitalWrite(SPI_CLK, LOW);
      }
  }
    
  //Set the Chips Select pin high to end the SPI packet.
  digitalWrite(CS, HIGH);
  digitalWrite(SPI_CLK, HIGH);
}

void writeRegister(char registerAddress, int value){  
  //Set the Chip select pin low to start an SPI packet.
  //printf("set CS LOW\n");		
  digitalWrite(CS, LOW);
  digitalWrite(SPI_CLK, LOW);
  
  //write WRITE - high
  char packet = 0x7f & registerAddress;
  
  //transfer WRITE + MB + address
  int i = 0;
  for (i = 0; i < 8; i++) {
    if (packet & 0x80) {

      digitalWrite(SPI_OUT, HIGH);
    }
    else {

      digitalWrite(SPI_OUT, LOW);
    }
    
    packet <<= 1;
    digitalWrite(SPI_CLK, HIGH);
    digitalWrite(SPI_CLK, LOW);
  }
    

  //transfer data

  i = 0;
  for (i = 0; i < 8; i++) {
        
    if (value & 0x80) {

      digitalWrite(SPI_OUT, HIGH);
    }
    else {

      digitalWrite(SPI_OUT, LOW);
    }
    
    value <<= 1;
    digitalWrite(SPI_CLK, HIGH);
    digitalWrite(SPI_CLK, LOW);
  }

    
  //Set the Chips Select pin high to end the SPI packet.
  digitalWrite(CS, HIGH);
  digitalWrite(SPI_CLK, HIGH);
}

int readBatchRegister(char registerAddress, int length){  
#ifdef SAFETY
  register unsigned int r11_val asm ("r4");
  unsigned char *tmp_reg_addr;
#endif
  //Set the Chip select pin low to start an SPI packet.
  //printf("Reading from %d\n", registerAddress);		
  //printf("set CS LOW\n");		
  digitalWrite(CS, LOW);
  digitalWrite(SPI_CLK, LOW);
  
  //write READ - high
  //set packet to READ (msb high)
  //printf("set READ HIGH\n");		
  char packet = 0x80 | registerAddress;
  
  //set packet 2nd MSB (MultiByte read) to HIGH
  packet = 0x40 | packet;
  
  //DEBUG PRINT

  
  //transfer READ + MB + address
  int i = 0;
  for (i = 0; i < 8; i++) {
    if (packet & 0x80) {

      digitalWrite(SPI_OUT, HIGH);
    }
    else {

      digitalWrite(SPI_OUT, LOW);
    }
    
    packet <<= 1;
    digitalWrite(SPI_CLK, HIGH);
    digitalWrite(SPI_CLK, LOW);
  }

  char adcount = 0;
  int j = 0;
  //read register data from DOUT
  digitalWrite(SPI_OUT, LOW);
  for(j = 0; j < length; j++) {
      for (i = 0; i < 8; i++) {
        adcount <<= 1;
        if (digitalRead(SPI_IN)) {
          adcount |= 0x1;

        }
        digitalWrite(SPI_CLK, HIGH);
        digitalWrite(SPI_CLK, LOW);
      }
#ifdef SAFETY
      tmp_reg_addr = (unsigned char *) &(regArray[j]);
      SAFETY_CHECK(tmp_reg_addr, r11_val, 3, *tmp_reg_addr = adcount);
#else
      regArray[j] = adcount;
#endif
  }

  
  //Set the Chips Select pin high to end the SPI packet.
  digitalWrite(CS, HIGH);
  digitalWrite(SPI_CLK, HIGH);
  return adcount;
}


int readRegister(char registerAddress){  
  //Set the Chip select pin low to start an SPI packet.
  //printf("Reading from %d\n", registerAddress);		
  //printf("set CS LOW\n");		
  digitalWrite(CS, LOW);
  digitalWrite(SPI_CLK, LOW);
  
  //write READ - high
  //set packet to READ (msb high)
  //printf("set READ HIGH\n");		
  char packet = 0x80 | registerAddress;
  
  //set packet 2nd MSB (MultiByte read) to HIGH
  packet = 0x40 | packet;
  
  //DEBUG PRINT

  
  //transfer READ + MB + address
  int i = 0;
  for (i = 0; i < 8; i++) {
    if (packet & 0x80) {

      digitalWrite(SPI_OUT, HIGH);
    }
    else {

      digitalWrite(SPI_OUT, LOW);
    }
    
    packet <<= 1;
    digitalWrite(SPI_CLK, HIGH);
    digitalWrite(SPI_CLK, LOW);
  }

  //read register data from DOUT
  int adcout = 0;
  digitalWrite(SPI_OUT, LOW);
  for (i = 0; i < 8; i++) {
    adcout <<= 1;
    if (digitalRead(SPI_IN)) {
      adcout |= 0x1;

    }
    digitalWrite(SPI_CLK, HIGH);
    digitalWrite(SPI_CLK, LOW);
  }

  
  //Set the Chips Select pin high to end the SPI packet.
  digitalWrite(CS, HIGH);
  digitalWrite(SPI_CLK, HIGH);
  return adcout;
}

int initSPI() {
  pinMode(SPI_OUT, OUTPUT);
  pinMode(SPI_IN, INPUT);
  pinMode(SPI_CLK, OUTPUT);
  pinMode(CS, OUTPUT);

  //setup
  digitalWrite(SPI_CLK, HIGH);  // start clock HIGH
  digitalWrite(CS, HIGH);     // bring CS HIGH
  digitalWrite(SPI_OUT, LOW);
    
    
}
