/* Test Program for AUSPICE-R
 * Authors: Jiaqi Tan <tanjiaqi@cmu.edu>, Hui Jun Tay <htay@andrew.cmu.edu>
 * Modifications Copyright (c) 2016 Jiaqi Tan, Hui Jun Tay
 * 15 June 2016
 * 
 * The AUSPICE-R Test Programs are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * Originally part of the WiringPi library (http://wiringpi.com)
 * Modified to add safety-checks for AUSPICE and AUSPICE-R
 */

/*
 * wiringPi:
 *	Arduino compatable (ish) Wiring library for the Raspberry Pi
 *	Copyright (c) 2012 Gordon Henderson
 ***********************************************************************
 * This file is part of wiringPi:
 *	https://projects.drogon.net/raspberry-pi/wiringpi/
 *
 *    wiringPi is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    wiringPi is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with wiringPi.  If not, see <http://www.gnu.org/licenses/>.
 ***********************************************************************
 */

#include <swi_functions.h
#include <rpi_defines.h>
#include <rpi.h>

#ifndef MIN_SAFE_MEM
#define MIN_SAFE_MEM 0xbd00
#endif

#ifdef SAFETY
#define WORD_SIZE 4
#define ABS_MAX_SAFE_MEM 0xBF000000
#define MAX_SAFE_MEM ((unsigned int) (ABS_MAX_SAFE_MEM - WORD_SIZE))
#define GET_FRAME_POINTER(dest_var)		\
  asm ( "mov r4,r11"				\
	: "=r" (dest_var)			\
	: /* no inputs */			\
	: /* no clobber */)			
#endif
char txt [11];

#define	TIMER_CONTROL	(0x408 >> 2)
#define	TIMER_PRE_DIV	(0x41C >> 2)
#define	TIMER_IRQ_RAW	(0x410 >> 2)

static volatile uint32_t *gpio;
static volatile uint32_t *pwm;
static volatile uint32_t *clk;
static volatile uint32_t *pads;
static volatile uint32_t *timer;
static volatile uint32_t *timerIrqRaw;

static volatile unsigned int RASPBERRY_PI_PERI_BASE;
static volatile unsigned int GPIO_PADS;
static volatile unsigned int GPIO_CLOCK_BASE;
static volatile unsigned int GPIO_BASE;
static volatile unsigned int GPIO_TIMER;
static volatile unsigned int GPIO_PWM;

static int wiringPiMode = WPI_MODE_UNINITIALISED ;
static int piModel2 = 0;
#define PI_REV 2
#define PAGE_SIZE 4


// gpioToGPFSEL:
//	Map a BCM_GPIO pin to it's Function Selection
//	control port. (GPFSEL 0-5)
//	Groups of 10 - 3 bits per Function - 30 bits per port

static uint8_t gpioToGPFSEL [] =
{
  0,0,0,0,0,0,0,0,0,0,
  1,1,1,1,1,1,1,1,1,1,
  2,2,2,2,2,2,2,2,2,2,
  3,3,3,3,3,3,3,3,3,3,
  4,4,4,4,4,4,4,4,4,4,
  5,5,5,5,5,5,5,5,5,5,
} ;

// gpioToShift
//	Define the shift up for the 3 bits per pin in each GPFSEL port

static uint8_t gpioToShift [] =
{
  0,3,6,9,12,15,18,21,24,27,
  0,3,6,9,12,15,18,21,24,27,
  0,3,6,9,12,15,18,21,24,27,
  0,3,6,9,12,15,18,21,24,27,
  0,3,6,9,12,15,18,21,24,27,
} ;

// gpioToGPSET:
//	(Word) offset to the GPIO Set registers for each GPIO pin

static uint8_t gpioToGPSET [] =
{
   7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
   8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
} ;

// gpioToGPCLR:
//	(Word) offset to the GPIO Clear registers for each GPIO pin

static uint8_t gpioToGPCLR [] =
{
  10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,
  11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,
} ;

// Doing it the Arduino way with lookup tables...
//	Yes, it's probably more innefficient than all the bit-twidling, but it
//	does tend to make it all a bit clearer. At least to me!

// pinToGpio:
//	Take a Wiring pin (0 through X) and re-map it to the BCM_GPIO pin
//	Cope for 3 different board revisions here.

static int *pinToGpio ;

// Revision 1, 1.1:

static int pinToGpioR1 [64] =
{
  17, 18, 21, 22, 23, 24, 25, 4,	// From the Original Wiki - GPIO 0 through 7:	wpi  0 -  7
   0,  1,				// I2C  - SDA1, SCL1				wpi  8 -  9
   8,  7,				// SPI  - CE1, CE0				wpi 10 - 11
  10,  9, 11, 				// SPI  - MOSI, MISO, SCLK			wpi 12 - 14
  14, 15,				// UART - Tx, Rx				wpi 15 - 16

// Padding:

      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,	// ... 31
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,	// ... 47
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,	// ... 63
} ;

// Revision 2:

static int pinToGpioR2 [64] =
{
  17, 18, 27, 22, 23, 24, 25, 4,	// From the Original Wiki - GPIO 0 through 7:	wpi  0 -  7
   2,  3,				// I2C  - SDA0, SCL0				wpi  8 -  9
   8,  7,				// SPI  - CE1, CE0				wpi 10 - 11
  10,  9, 11, 				// SPI  - MOSI, MISO, SCLK			wpi 12 - 14
  14, 15,				// UART - Tx, Rx				wpi 15 - 16
  28, 29, 30, 31,			// Rev 2: New GPIOs 8 though 11			wpi 17 - 20
   5,  6, 13, 19, 26,			// B+						wpi 21, 22, 23, 24, 25
  12, 16, 20, 21,			// B+						wpi 26, 27, 28, 29
   0,  1,				// B+						wpi 30, 31

// Padding:

  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,	// ... 47
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,	// ... 63
} ;

// physToGpio:
//	Take a physical pin (1 through 26) and re-map it to the BCM_GPIO pin
//	Cope for 2 different board revisions here.
//	Also add in the P5 connector, so the P5 pins are 3,4,5,6, so 53,54,55,56

static int *physToGpio ;

static int physToGpioR1 [64] =
{
  -1,		// 0
  -1, -1,	// 1, 2
   0, -1,
   1, -1,
   4, 14,
  -1, 15,
  17, 18,
  21, -1,
  22, 23,
  -1, 24,
  10, -1,
   9, 25,
  11,  8,
  -1,  7,	// 25, 26

                                              -1, -1, -1, -1, -1,	// ... 31
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,	// ... 47
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,	// ... 63
} ;

static int physToGpioR2 [64] =
{
  -1,		// 0
  -1, -1,	// 1, 2
   2, -1,
   3, -1,
   4, 14,
  -1, 15,
  17, 18,
  27, -1,
  22, 23,
  -1, 24,
  10, -1,
   9, 25,
  11,  8,
  -1,  7,	// 25, 26

// B+

   0,  1,
   5, -1,
   6, 12,
  13, -1,
  19, 16,
  26, 20,
  -1, 21,

// the P5 connector on the Rev 2 boards:

  -1, -1,
  -1, -1,
  -1, -1,
  -1, -1,
  -1, -1,
  28, 29,
  30, 31,
  -1, -1,
  -1, -1,
  -1, -1,
  -1, -1,
} ;

// sysFds:
//	Map a file descriptor from the /sys/class/gpio/gpioX/value

static int sysFds [64] =
{
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
} ;

// gpioToGPLEV:
//	(Word) offset to the GPIO Input level registers for each GPIO pin

static uint8_t gpioToGPLEV [] =
{
  13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,
  14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,
} ;


// GPPUD:
//	GPIO Pin pull up/down register

#define	GPPUD	37

// gpioToPUDCLK
//	(Word) offset to the Pull Up Down Clock regsiter

static uint8_t gpioToPUDCLK [] =
{
  38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,38,
  39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,39,
} ;


//FUNCTIONS

void delay (unsigned int howLong)
{
  struct timespec sleeper, dummy ;
  sleeper.tv_sec  = (time_t)(howLong / 1000) ;
  sleeper.tv_nsec = (long)(howLong % 1000) * 1000000 ;
  c_nanosleep (&sleeper, &dummy) ;
}

void delayMicroseconds (unsigned int howLong)
{
  struct timespec sleeper ;
  unsigned int uSecs = howLong % 1000000 ;
  unsigned int wSecs = howLong / 1000000 ;

  sleeper.tv_sec  = wSecs ;
  sleeper.tv_nsec = (long)(uSecs * 1000L) ;
  c_nanosleep (&sleeper, 0) ;
}

int setup_piModel1_rev2() {
  int fd, i, retval;
  uint32_t *tmp_ptr;
#define STACK_SIZE 3
#ifdef SAFETY
  register unsigned int r11_val asm ("r4");
#endif
 
  fd = c_open("/dev/mem", O_RDWR | O_SYNC | O_CLOEXEC);
  if (fd < 0) { return -1; }

#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&RASPBERRY_PI_PERI_BASE) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    RASPBERRY_PI_PERI_BASE = 0x20000000 ; // this was confirmed against actual wiringPi
#ifdef SAFETY
  }
#endif
#ifdef SAFETY  
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&GPIO_PADS) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    GPIO_PADS 	  = RASPBERRY_PI_PERI_BASE + 0x00100000 ;
#ifdef SAFETY
  }
#endif
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&GPIO_CLOCK_BASE) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    GPIO_CLOCK_BASE = RASPBERRY_PI_PERI_BASE + 0x00101000 ;
#ifdef SAFETY
  }
#endif
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&GPIO_BASE) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    GPIO_BASE	  = RASPBERRY_PI_PERI_BASE + 0x00200000 ;
#ifdef SAFETY
  }
#endif
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&GPIO_TIMER) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    GPIO_TIMER	  = RASPBERRY_PI_PERI_BASE + 0x0000B000 ;
#ifdef SAFETY
  }
#endif
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&GPIO_PWM) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    GPIO_PWM	  = RASPBERRY_PI_PERI_BASE + 0x0020C000 ;
#ifdef SAFETY
  }
#endif
  tmp_ptr = (uint32_t *) c_mmap(0, BLOCK_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, GPIO_BASE);
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&gpio) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    gpio = tmp_ptr;
#ifdef SAFETY
  }
#endif

  tmp_ptr = (uint32_t *) c_mmap(0, BLOCK_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, GPIO_PWM);
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&pwm) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    pwm = tmp_ptr;
#ifdef SAFETY
  }
#endif

  tmp_ptr = (uint32_t *) c_mmap(0, BLOCK_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, GPIO_CLOCK_BASE);
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&clk) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    clk = tmp_ptr;
#ifdef SAFETY
  }
#endif

  tmp_ptr = (uint32_t *) c_mmap(0, BLOCK_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, GPIO_PADS);
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&pads) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    pads = tmp_ptr;
#ifdef SAFETY
  }
#endif

//ADD TIMER
  
  tmp_ptr = (uint32_t *) c_mmap(0, BLOCK_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, GPIO_TIMER);
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&timer) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    timer = tmp_ptr;
#ifdef SAFETY
  }
#endif

#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&wiringPiMode) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    wiringPiMode = WPI_MODE_PINS;
#ifdef SAFETY
  }
#endif

#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&pinToGpio) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    pinToGpio =  pinToGpioR2 ;
#ifdef SAFETY
  }
#endif
#ifdef SAFETY
  GET_FRAME_POINTER(r11_val);
  if (((unsigned int)(&physToGpio) < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
      (r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
    physToGpio = physToGpioR2 ;
#ifdef SAFETY
  }
#endif  
#undef STACK_SIZE
  if ((((int32_t) gpio) == -1) || (((int32_t) pwm) == -1) || 
      (((int32_t) clk) == -1) || (((int32_t) pads) == -1) || ((int32_t) timer == -1)) {
    return -1;
  } else {
    return 0;
  }
}

//int setup_piModel2_rev2() {
//  return -1;
//}

//timer function

uint32_t getSysTicks() {    
  unsigned int *time = (unsigned int *)((char *)timer + 4);
  return time[0];
}

void pinMode (int pin, int mode)
{    
  volatile uint32_t *my_gpio;
  int fSel, shift, i ;
  int origPin = pin;
#ifdef SAFETY
  register unsigned int r11_val asm ("r4");
#endif

  if (wiringPiMode == WPI_MODE_PINS)
    pin = pinToGpio [pin] ;
  else if (wiringPiMode == WPI_MODE_PHYS)
    pin = physToGpio [pin] ;
  else if (wiringPiMode != WPI_MODE_GPIO)
    return ;

  fSel    = gpioToGPFSEL [pin] ;
  shift   = gpioToShift  [pin] ;

  if (mode == INPUT) {
	  my_gpio = gpio + fSel;
#ifdef SAFETY
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int) my_gpio >= MIN_SAFE_MEM) &&
	((unsigned int) my_gpio <= MAX_SAFE_MEM) && 
	((unsigned int) my_gpio < (r11_val - (3 * WORD_SIZE))) && 
	(r11_val >= (3 * WORD_SIZE))) { 
#endif
      *my_gpio = (*(gpio + fSel) & ~(7 << shift)) ; // Sets bits to zero = input
#ifdef SAFETY
    }
#endif
  } else if (mode == OUTPUT) {
	  my_gpio = gpio + fSel;
#ifdef SAFETY
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int) my_gpio >= MIN_SAFE_MEM) &&
	((unsigned int) my_gpio <= MAX_SAFE_MEM) && 
	((unsigned int) my_gpio < (r11_val - (3 * WORD_SIZE))) && 
	(r11_val >= (3 * WORD_SIZE))) { 
#endif
      *my_gpio = (*(gpio + fSel) & ~(7 << shift)) | (1 << shift) ;
#ifdef SAFETY
    }
#endif
  }
  // }
}

//WRITE

void digitalWrite (int pin, int value)
{
  int i;
  volatile uint32_t *my_gpio;
#ifdef SAFETY
  register unsigned int r11_val asm ("r4");
#endif

  pin = pinToGpio [pin] ;
  
  if (value == LOW) {
	  my_gpio = gpio + gpioToGPCLR [pin];
#ifdef SAFETY
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int) my_gpio >= MIN_SAFE_MEM) &&
	((unsigned int) my_gpio <= MAX_SAFE_MEM) && 
	((unsigned int) my_gpio < (r11_val - (2 * WORD_SIZE))) && 
	(r11_val >= (2 * WORD_SIZE))) { 
#endif
      *my_gpio = (1 << (pin & 31)) ;
#ifdef SAFETY
    }
#endif
  }
  else {
	  my_gpio = gpio + gpioToGPSET [pin];
#ifdef SAFETY
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int) my_gpio >= MIN_SAFE_MEM) &&
	((unsigned int) my_gpio <= MAX_SAFE_MEM) && 
	((unsigned int) my_gpio < (r11_val - (2 * WORD_SIZE))) && 
	(r11_val >= (2 * WORD_SIZE))) { 
#endif
      *my_gpio = (1 << (pin & 31)) ;
#ifdef SAFETY
    }
#endif
  }
}

void pullUpDnControl (int pin, int pud) 
{
#ifdef SAFETY
  register unsigned int r11_val asm ("r4");
#endif
  volatile uint32_t *my_gpio;
#define STACK_SIZE 3

  if ((pin & PI_GPIO_MASK) == 0)		// On-Board Pin
  {

    my_gpio = gpio + GPPUD;
#ifdef SAFETY
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int) my_gpio >= MIN_SAFE_MEM) &&
	((unsigned int) my_gpio <= MAX_SAFE_MEM) && 
	((unsigned int) my_gpio < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
	(r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
      *my_gpio              = pud & 3 ;		
#ifdef SAFETY
    }
#endif
    delay(1) ;

    my_gpio = gpio + gpioToPUDCLK [pin];
#ifdef SAFETY
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int) my_gpio >= MIN_SAFE_MEM) &&
	((unsigned int) my_gpio <= MAX_SAFE_MEM) && 
	((unsigned int) my_gpio < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
	(r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
      *(my_gpio) = 1 << (pin & 31) ;	
#ifdef SAFETY
    }
#endif
    delay(1) ;
    
    my_gpio = gpio + GPPUD;
#ifdef SAFETY
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int) my_gpio >= MIN_SAFE_MEM) &&
	((unsigned int) my_gpio <= MAX_SAFE_MEM) && 
	((unsigned int) my_gpio < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
	(r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
      *my_gpio              = 0 ;			
#ifdef SAFETY
    }
#endif
    delay(1) ;
    
    my_gpio = gpio + gpioToPUDCLK [pin];
#ifdef SAFETY
    GET_FRAME_POINTER(r11_val);
    if (((unsigned int) my_gpio >= MIN_SAFE_MEM) &&
	((unsigned int) my_gpio <= MAX_SAFE_MEM) && 
	((unsigned int) my_gpio < (r11_val - (STACK_SIZE * WORD_SIZE))) && 
	(r11_val >= (STACK_SIZE * WORD_SIZE))) { 
#endif
      *my_gpio = 0 ;			
#ifdef SAFETY
    }
#endif
    delay(1) ;
  }
#undef STACK_SIZE
}

int digitalRead (int pin)
{
  if ((pin & PI_GPIO_MASK) == 0) {	// On-Board Pin
      pin = pinToGpio [pin] ;
  
    int val = (*(gpio + gpioToGPLEV [pin]) & (1 << (pin & 31)));
    
    if ((*(gpio + gpioToGPLEV [pin]) & (1 << (pin & 31))) != 0)
      return HIGH ;
    else
      return LOW ;
  }

}
