/*
 * network type definitions
 */

// from bionic/libc/arch-arm/include/machine/_types.h

#ifndef __TYPES_H_
#define __TYPES_H_


/* bsd */
typedef unsigned char		u_char;
typedef unsigned short		u_short;
typedef unsigned int		u_int;
typedef unsigned long		u_long;

#ifndef __ASSEMBLY__
/*
 * __xx is ok: it doesn't pollute the POSIX namespace. Use these in the
 * header files exported to user space
 */

typedef __signed__ char __s8;
typedef unsigned char __u8;

typedef __signed__ short __s16;
typedef unsigned short __u16;

typedef __signed__ int __s32;
typedef unsigned int __u32;

#ifdef __GNUC__
__extension__ typedef __signed__ long long __s64;
__extension__ typedef unsigned long long __u64;
#else
typedef __signed__ long long __s64;
typedef unsigned long long __u64;
#endif

#endif /* __ASSEMBLY__ */

#if !defined(__ARM_EABI__)
/* the kernel defines size_t as unsigned int, but g++ wants it to be unsigned long */
#define _SIZE_T
#define _SSIZE_T
#define _PTRDIFF_T
typedef unsigned long  size_t;
typedef long           ssize_t;
typedef long           ptrdiff_t;
#endif

/* 7.18.1.1 Exact-width integer types */
typedef	__signed char		__int8_t;
typedef	unsigned char		__uint8_t;
typedef	short			__int16_t;
typedef	unsigned short		__uint16_t;
typedef	int			__int32_t;
typedef	unsigned int		__uint32_t;
/* LONGLONG */
typedef	long long		__int64_t;
/* LONGLONG */
typedef	unsigned long long	__uint64_t;

//#endif

typedef __uint16_t	__sa_family_t; // sys/_types.h
typedef __sa_family_t sa_family_t; // sys/socket.h

#define __bitwise 
// linux/types.h
typedef		__u16		u_int16_t;
typedef		__u32		u_int32_t;
typedef	__u32 u_int32_t;
typedef __u16 __bitwise __be16;
typedef __u32 __bitwise __be32;
// linux/in.h
typedef __u16 __bitwise __be16;
typedef __u32 __bitwise __be32;

// sys/socket.h
typedef int socklen_t;

#define SOCK_STREAM      1
#define SOCK_DGRAM       2
#define SOCK_RAW         3
#define SOCK_RDM         4
#define SOCK_SEQPACKET   5
#define SOCK_PACKET      10

#define AF_UNSPEC 0
#define AF_UNIX 1
#define AF_LOCAL 1
#define AF_INET 2
#define AF_AX25 3
#define AF_IPX 4
#define AF_APPLETALK 5
#define AF_NETROM 6
#define AF_BRIDGE 7
#define AF_ATMPVC 8
#define AF_X25 9
#define AF_INET6 10
#define AF_ROSE 11
#define AF_DECnet 12
#define AF_NETBEUI 13
#define AF_SECURITY 14
#define AF_KEY 15
#define AF_NETLINK 16
#define AF_ROUTE AF_NETLINK
#define AF_PACKET 17
#define AF_ASH 18
#define AF_ECONET 19
#define AF_ATMSVC 20
#define AF_SNA 22
#define AF_IRDA 23
#define AF_PPPOX 24
#define AF_WANPIPE 25
#define AF_LLC 26
#define AF_TIPC 30
#define AF_BLUETOOTH 31
#define AF_CAIF 38
#define AF_MAX 39

// sys/socket.h
struct sockaddr {
 sa_family_t sa_family;
 char sa_data[14];
};

struct in_addr {
  __be32	s_addr;
};

#define __SOCK_SIZE__	16		/* sizeof(struct sockaddr)	*/
struct sockaddr_in {
  sa_family_t		sin_family;	/* Address family		*/
  __be16		sin_port;	/* Port number			*/
  struct in_addr	sin_addr;	/* Internet address		*/

  /* Pad to size of `struct sockaddr'. */
  unsigned char		__pad[__SOCK_SIZE__ - sizeof(short int) -
			sizeof(unsigned short int) - sizeof(struct in_addr)];
};

// netdb.h
struct hostent {
  char	*h_name;	/* official name of host */
  char	**h_aliases;	/* alias list */
  int	h_addrtype;	/* host address type */
  int	h_length;	/* length of address */
  char	**h_addr_list;	/* list of addresses from name server */
#define	h_addr	h_addr_list[0]	/* address, for backward compatibility */
};


// netinet/in.h

// typedef signed int ssize_t;
// typedef unsigned int size_t;
// typedef unsigned int off_t;

/*
 * time.h definitions
 */

/* from bits/types.h */
#define __SWORD_TYPE           int
#define __SLONGWORD_TYPE        long int

/* from bits/typesizes.h */
#define __TIME_T_TYPE           __SLONGWORD_TYPE
#define __OFF_T_TYPE            __SLONGWORD_TYPE
#define __SSIZE_T_TYPE          __SWORD_TYPE

#ifndef __SIZE_TYPE
#define __SIZE_TYPE__ long unsigned int /* lib/gcc/arm-linux-gnueabihf/4.6/include/stddef.h */
#endif

typedef __OFF_T_TYPE __off_t;
typedef __off_t off_t;
typedef __SSIZE_T_TYPE ssize_t; /* Type of a byte count, or error.  */
typedef __SIZE_TYPE__ size_t; /* glob.h */
typedef __TIME_T_TYPE __time_t;
typedef __time_t time_t;

/* from /usr/lib/gcc/arm-linux-gnueabihf/4.6/include */
// typedef long unsigned int size_t;
// typedef long ssize_t;

/* from /usr/include/time.h */

struct timespec
  {
    __time_t tv_sec;            /* Seconds.  */
    long int tv_nsec;           /* Nanoseconds.  */
  };

#endif // __TYPES_H_
