#!/usr/bin/env python
#
######################################################################
# Test Program for AUSPICE-R
# Author: Jiaqi Tan <tanjiaqi@cmu.edu>
# Copyright (c) 2016 Jiaqi Tan
# 15 June 2016
# 
# The AUSPICE-R Test Programs are distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
######################################################################
 
import io
import sys
import re

if (len(sys.argv) < 2):
    raise Exception("Usage: "+sys.argv[0]+" [input file]")
# end if

INFILENAME = sys.argv[1]
INFILE = io.open(INFILENAME, "rb+")

text_header_pat = re.compile(r"Disassembly of section \.text")
rodata_header_pat = re.compile(r"Disassembly of section \.rodata")
header_pat = re.compile(r"Disassembly of section")
instr_whitespace_pat = re.compile(r"^\s+")
addr_code_pat = re.compile(r"^\s+([0-9abcdef]+):\s+([0-9abcdef]+)")
func_header_pat = re.compile(r"^([0-9abcdef]+) <([a-zA-Z0-9_]+)+>:")

in_text = False
in_data = False
fun_list = []
fun_addr_list = []
fun_end_addr_list = []
last_func_addr = -1
last_addr = -1
last_code_addr = -1
for line0 in INFILE:
    line = line0.rstrip()

    if (in_text):
        if (instr_whitespace_pat.match(line) != None):
            addr_code_matcher = addr_code_pat.match(line)
            if (addr_code_matcher != None):
                addr = addr_code_matcher.group(1).upper()
                code = addr_code_matcher.group(2)
                last_addr = addr;
                last_code_addr = addr;
                print(addr+","+code)
            # end if
        elif (func_header_pat.match(line) != None):
            func_header_matcher = func_header_pat.match(line)
            if (func_header_matcher != None):
                addr = func_header_matcher.group(1).upper().lstrip("0")
                if (last_func_addr > 0):
                    last_func_addr = (hex(int(addr,16)-4)).upper().lstrip("0X")
                    fun_end_addr_list.append(last_func_addr)
                last_func_addr = addr
                func_name = func_header_matcher.group(2)
                fun_list.append(func_name)
                fun_addr_list.append(addr)
            # end if
        # end if
    elif (in_data):
        if (instr_whitespace_pat.match(line) != None):
            addr_code_matcher = addr_code_pat.match(line)
            if (addr_code_matcher != None):
                addr = addr_code_matcher.group(1).upper()
                code = addr_code_matcher.group(2)
                print(addr+","+code)
            # end if
        # end if
    # end if

    if ( (not in_text) and (text_header_pat.match(line) != None) ):
        in_text = True
        print(".text");
    elif ( (not in_data) and (rodata_header_pat.match(line) != None) ):
        in_data = True
        in_text = False
        print(".rodata");
    elif (header_pat.match(line) != None):
        in_data = False
        in_text = False
    # end if
# end for
fun_end_addr_list.append(last_code_addr)

print(".functions")
i = 0
for func in fun_list:
    curr_func_addr = fun_addr_list[i]
    curr_func_end_addr = fun_end_addr_list[i]
    if (func[0] == '_'):
        func = "x" + func
    print(curr_func_addr+","+curr_func_end_addr+","+func)
    i = i + 1
